'''
This module contains different variations of the Physics Informer Neural Network model using the JaxModel API
'''
import numpy as np
import time
from typing import Any, Callable, Iterable, List, Optional, Tuple, Union

from collections.abc import Sequence as SequenceCollection

from deepchem.data import Dataset
from deepchem.models.losses import Loss
from deepchem.models.optimizers import Optimizer
from deepchem.utils.typing import LossFn, OneOrMany
from deepchem.trans.transformers import Transformer, undo_transforms

# JAX dependencies
import jax
import jax.numpy as jnp
import haiku as hk
import optax
from deepchem.models.jax_models.jax_model import JaxModel
from deepchem.models.jax_models.jax_model import create_default_gradient_fn, create_default_eval_fn

import logging
import warnings

logger = logging.getLogger(__name__)


def create_default_update_fn(optimizer: optax.GradientTransformation,
                             model_loss: Callable):
  """
  This function calls the update function, to implement the backpropagation
  """

  @jax.jit
  def update(params, opt_state, batch, target, weights,
             rng) -> Tuple[hk.Params, optax.OptState, jnp.ndarray]:
    batch_loss, grads = jax.value_and_grad(model_loss)(params, target, weights,
                                                       rng, *batch)
    updates, opt_state = optimizer.update(grads, opt_state)
    new_params = optax.apply_updates(params, updates)
    return new_params, opt_state, batch_loss

  return update


class PINNModel(JaxModel):
  """
  This is class is derived from the JaxModel class and methods are also very similar to JaxModel,
  but it has the option of passing multiple arguments(Done using *args) suitable for PINNs model.
  Ex - Approximating f(x, y, z, t) satisfying a Linear differential equation.

  This model is recommended for linear partial differential equations but if you can accurately write
  the gradient function in Jax depending on your use case, then it will work as well.

  This class requires two functions apart from the usual function definition and weights

  [1] **grad_fn** : Each PINNs have a different strategy for calculating its final losses. This
  function tells the PINNModel how to go about computing the derivatives for backpropagation.
  It should follow this format:

  >>>
  >> def gradient_fn(forward_fn, loss_outputs, initial_data):
  >>
  >>  def model_loss(params, target, weights, rng, ...):
  >>
  >>    # write code using the arguments.
  >>    # ... indicates the variable number of positional arguments.
  >>    return
  >>
  >>  return model_loss

  "..." can be replaced with various arguments like (x, y, z, y) but should match with eval_fn

  [2] **eval_fn**: Function for defining how the model needs to compute during inference.
  It should follow this format

  >>>
  >> def create_eval_fn(forward_fn, params):
  >>  def eval_model(..., rng=None):
  >>    # write code here using arguments
  >>
  >>    return
  >>  return eval_model

  "..." can be replaced with various arguments like (x, y, z, y) but should match with grad_fn

  [3] boundary_data:
  For a detailed example, check out - deepchem/models/jax_models/tests/test_pinn.py where we have
  solved f'(x) = -sin(x)

  References
  ----------
  .. [1] Raissi et. al. "Physics-informed neural networks: A deep learning framework for solving
     forward and inverse problems involving nonlinear partial differential equations" Journal of
     Computational Physics https://doi.org/10.1016/j.jcp.2018.10.045

  .. [2] Raissi et. al. "Physics Informed Deep Learning (Part I): Data-driven
     Solutions of Nonlinear Partial Differential Equations" arXiv preprint arXiv:1711.10561

  Notes
  -----
  This class requires Jax, Haiku and Optax to be installed.
  """

  def __init__(self,
               forward_fn: hk.State,
               params: hk.Params,
               initial_data: dict = {},
               output_types: Optional[List[str]] = None,
               batch_size: int = 100,
               learning_rate: float = 0.001,
               optimizer: Union[optax.GradientTransformation, Optimizer] = None,
               grad_fn: Callable = create_default_gradient_fn,
               update_fn: Callable = create_default_update_fn,
               eval_fn: Callable = create_default_eval_fn,
               rng=jax.random.PRNGKey(1),
               log_frequency: int = 100,
               **kwargs):
    """
    Parameters
    ----------
    forward_fn: hk.State or Function
      Any Jax based model that has a `apply` method for computing the network. Currently
      only haiku models are supported.
    params: hk.Params
      The parameter of the Jax based networks
    initial_data: dict
      This acts as a session variable which will be passed as a dictionary in grad_fn
    output_types: list of strings, optional (default None)
      the type of each output from the model, as described above
    batch_size: int, optional (default 100)
      default batch size for training and evaluating
    learning_rate: float or LearningRateSchedule, optional (default 0.001)
      the learning rate to use for fitting.  If optimizer is specified, this is
      ignored.
    optimizer: optax object
      For the time being, it is optax object
    grad_fn: Callable (default create_default_gradient_fn)
      It defines how the loss function and gradients need to be calculated for the PINNs model
    update_fn: Callable (default create_default_update_fn)
      It defines how the weights need to be updated using backpropogation. We have used optax library
      for optimisation operations. Its reccomended to leave this default.
    eval_fn: Callable (default create_default_eval_fn)
      Function for defining on how the model needs to compute during inference.
    rng: jax.random.PRNGKey, optional (default 1)
      A default global PRNG key to use for drawing random numbers.
    log_frequency: int, optional (default 100)
      The frequency at which to log data. Data is logged using
      `logging` by default.
    """
    warnings.warn(
        'PinnModel is still in active development and we could change the design of the API in the future.'
    )
    self.boundary_data = initial_data
    super(PINNModel,
          self).__init__(forward_fn, params, None, output_types, batch_size,
                         learning_rate, optimizer, grad_fn, update_fn, eval_fn,
                         rng, log_frequency, **kwargs)

  def fit_generator(self,
                    generator: Iterable[Tuple[Any, Any, Any]],
                    loss: Union[Loss, LossFn] = None,
                    callbacks: Union[Callable, List[Callable]] = [],
                    all_losses: Optional[List[float]] = None) -> float:
    if not isinstance(callbacks, SequenceCollection):
      callbacks = [callbacks]
    self._ensure_built()
    avg_loss = 0.0
    last_avg_loss = 0.0
    averaged_batches = 0
    if loss is None:
      loss = self._loss_fn
    model_loss_fn = self._create_gradient_fn(self.forward_fn,
                                             self._loss_outputs,
                                             self.boundary_data)
    grad_update = self._create_update_fn(self.optimizer, model_loss_fn)

    params, opt_state = self._get_trainable_params()
    rng = self.rng
    time1 = time.time()

    # Main training loop

    for batch in generator:
      inputs, labels, weights = self._prepare_batch(batch)

      if isinstance(inputs, list) and len(inputs) == 1:
        inputs = inputs[0]

      if isinstance(labels, list) and len(labels) == 1:
        labels = labels[0]

      if isinstance(weights, list) and len(weights) == 1:
        weights = weights[0]

      params, opt_state, batch_loss = grad_update(params,
                                                  opt_state,
                                                  inputs,
                                                  labels,
                                                  weights,
                                                  rng=rng)
      rng, _ = jax.random.split(rng)
      avg_loss += jax.device_get(batch_loss)
      self._global_step += 1
      current_step = self._global_step
      averaged_batches += 1
      should_log = (current_step % self.log_frequency == 0)

      if should_log:
        avg_loss = float(avg_loss) / averaged_batches
        logger.info('Ending global_step %d: Average loss %g' %
                    (current_step, avg_loss))
        if all_losses is not None:
          all_losses.append(avg_loss)
        # Capture the last avg_loss in case of return since we're resetting to 0 now
        last_avg_loss = avg_loss
        avg_loss = 0.0
        averaged_batches = 0
      for c in callbacks:
        c(self, current_step)

    # Report final results.
    if averaged_batches > 0:
      avg_loss = float(avg_loss) / averaged_batches
      logger.info('Ending global_step %d: Average loss %g' %
                  (current_step, avg_loss))
      if all_losses is not None:
        all_losses.append(avg_loss)
      last_avg_loss = avg_loss

    time2 = time.time()
    logger.info("TIMING: model fitting took %0.3f s" % (time2 - time1))
    self._set_trainable_params(params, opt_state)
    return last_avg_loss

  def _prepare_batch(self, batch):
    inputs, labels, weights = batch
    inputs = [
        x.astype(np.float32) if x.dtype == np.float64 else x for x in inputs
    ]
    inputs = [np.split(x, x.shape[1], 1) for x in inputs]
    if labels is not None:
      labels = [
          x.astype(np.float32) if x.dtype == np.float64 else x for x in labels
      ]
    else:
      labels = []

    if weights is not None:
      weights = [
          x.astype(np.float32) if x.dtype == np.float64 else x for x in weights
      ]
    else:
      weights = []

    return (inputs, labels, weights)

  def default_generator(
      self,
      dataset: Dataset,
      epochs: int = 1,
      mode: str = 'fit',
      deterministic: bool = True,
      pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
    """Create a generator that iterates batches for a dataset.
    Subclasses may override this method to customize how model inputs are
    generated from the data.

    Parameters
    ----------
    dataset: Dataset
      the data to iterate
    epochs: int
      the number of times to iterate over the full dataset
    mode: str
      allowed values are 'fit' (called during training), 'predict' (called
      during prediction), and 'uncertainty' (called during uncertainty
      prediction)
    deterministic: bool
      whether to iterate over the dataset in order, or randomly shuffle the
      data for each epoch
    pad_batches: bool
      whether to pad each batch up to this model's preferred batch size
    Returns
    -------
    a generator that iterates batches, each represented as a tuple of lists:
    ([inputs], [outputs], [weights])
    """

    for epoch in range(epochs):
      for (X_b, y_b, w_b, _) in dataset.iterbatches(batch_size=self.batch_size,
                                                    deterministic=deterministic,
                                                    pad_batches=pad_batches):
        yield ([X_b], [y_b], [w_b])

  def _predict(self, generator: Iterable[Tuple[Any, Any, Any]],
               transformers: List[Transformer], uncertainty: bool,
               other_output_types: Optional[OneOrMany[str]]):
    """
    Predict outputs for data provided by a generator.
    This is the private implementation of prediction.  Do not
    call it directly. Instead call one of the public prediction
    methods.
    Parameters
    ----------
    generator: generator
      this should generate batches, each represented as a tuple of the form
      (inputs, labels, weights).
    transformers: List[dc.trans.Transformers]
      Transformers that the input data has been transformed by.  The output
      is passed through these transformers to undo the transformations.
    uncertainty: bool
      specifies whether this is being called as part of estimating uncertainty.
      If True, it sets the training flag so that dropout will be enabled, and
      returns the values of the uncertainty outputs.
    other_output_types: list, optional
      Provides a list of other output_types (strings) to predict from model.
    Returns
    -------
    A NumpyArray if the model produces a single output, or a list of arrays otherwise.
    """
    results: Optional[List[List[np.ndarray]]] = None
    variances: Optional[List[List[np.ndarray]]] = None
    if uncertainty and (other_output_types is not None):
      raise ValueError(
          'This model cannot compute uncertainties and other output types simultaneously. Please invoke one at a time.'
      )
    if uncertainty:
      if self._variance_outputs is None or len(self._variance_outputs) == 0:
        raise ValueError('This model cannot compute uncertainties')
      if len(self._variance_outputs) != len(self._prediction_outputs):
        raise ValueError(
            'The number of variances must exactly match the number of outputs')
    if other_output_types:
      if self._other_outputs is None or len(self._other_outputs) == 0:
        raise ValueError(
            'This model cannot compute other outputs since no other output_types were specified.'
        )
    self._ensure_built()
    eval_fn = self._create_eval_fn(self.forward_fn, self.params)
    rng = self.rng

    for batch in generator:
      inputs, _, _ = self._prepare_batch(batch)

      if isinstance(inputs, list) and len(inputs) == 1:
        inputs = inputs[0]

      output_values = eval_fn(*inputs, rng)
      if isinstance(output_values, jnp.ndarray):
        output_values = [output_values]
      output_values = [jax.device_get(t) for t in output_values]

      # Apply tranformers and record results.
      if uncertainty:
        var = [output_values[i] for i in self._variance_outputs]
        if variances is None:
          variances = [var]
        else:
          for i, t in enumerate(var):
            variances[i].append(t)

      access_values = []
      if other_output_types:
        access_values += self._other_outputs
      elif self._prediction_outputs is not None:
        access_values += self._prediction_outputs

      if len(access_values) > 0:
        output_values = [output_values[i] for i in access_values]

      if len(transformers) > 0:
        if len(output_values) > 1:
          raise ValueError(
              "predict() does not support Transformers for models with multiple outputs."
          )
        elif len(output_values) == 1:
          output_values = [undo_transforms(output_values[0], transformers)]

      if results is None:
        results = [[] for i in range(len(output_values))]
      for i, t in enumerate(output_values):
        results[i].append(t)

    # Concatenate arrays to create the final results.
    final_results = []
    final_variances = []
    if results is not None:
      for r in results:
        final_results.append(np.concatenate(r, axis=0))
    if uncertainty and variances is not None:
      for v in variances:
        final_variances.append(np.concatenate(v, axis=0))
      return zip(final_results, final_variances)
    if len(final_results) == 1:
      return final_results[0]
    else:
      return final_results



"""
Contains class for random hyperparameter optimizations.
"""
import numpy as np
import os
import tempfile
import collections
import logging
import itertools
from typing import Dict, List, Optional, Tuple, Any, Callable

from deepchem.data import Dataset
from deepchem.trans import Transformer
from deepchem.models import Model
from deepchem.metrics import Metric
from deepchem.hyper.base_classes import HyperparamOpt
from deepchem.hyper.base_classes import _convert_hyperparam_dict_to_filename

logger = logging.getLogger(__name__)


class RandomHyperparamOpt(HyperparamOpt):
  """
  Provides simple random hyperparameter search capabilities.

  This class performs a random hyperparameter search over the specified
  hyperparameter space. This implementation is simple and samples
  hyperparameters from the hyperparameter space passed. However it does not
  use parallelization to speed up the search.

  Examples
  --------
  This example shows the type of constructor function expects.

  >>> import sklearn
  >>> import deepchem as dc
  >>> optimizer = dc.hyper.RandomHyperparamOpt(lambda **p: dc.models.GraphConvModel(**p), max_iter=5)

  Here's a more sophisticated example that shows how to optimize only
  some parameters of a model. In this case, we have some parameters we
  want to optimize, and others which we don't. To handle this type of
  search, we create a `model_builder` which hard codes some arguments.

  >>> import deepchem as dc
  >>> import numpy as np
  >>> from sklearn.linear_model import LogisticRegression as LR
  >>> # generating data
  >>> X = np.arange(1, 11, 1).reshape(-1, 1)
  >>> y = np.hstack((np.zeros(5), np.ones(5)))
  >>> dataset = dc.data.NumpyDataset(X, y)
  >>> # splitting dataset into train and test
  >>> splitter = dc.splits.RandomSplitter()
  >>> train_dataset, test_dataset = splitter.train_test_split(dataset)
  >>> # metric to evaluate result of a set of parameters
  >>> metric = dc.metrics.Metric(dc.metrics.accuracy_score)
  >>> # defining `model_builder`
  >>> def model_builder(**model_params):
  ...   penalty = model_params['penalty']
  ...   solver = model_params['solver']
  ...   lr = LR(penalty=penalty, solver=solver, max_iter=100)
  ...   return dc.models.SklearnModel(lr)
  >>> # the parameters which are to be optimized
  >>> params = {
  ...   'penalty': ['l1', 'l2'],
  ...   'solver': ['liblinear', 'saga']
  ...   }
  >>> # Creating optimizer and searching over hyperparameters
  >>> optimizer = dc.hyper.RandomHyperparamOpt(model_builder, max_iter=100)
  >>> best_model, best_hyperparams, all_results = optimizer.hyperparam_search(params, train_dataset, test_dataset, metric)
  >>> best_hyperparams['penalty'], best_hyperparams['solver']  # doctest: +SKIP
  ('l2', 'saga')

  Parameters
  ----------
  model_builder: constructor function.
    This parameter must be constructor function which returns an
    object which is an instance of `dc.models.Model`. This function
    must accept two arguments, `model_params` of type `dict` and
    `model_dir`, a string specifying a path to a model directory.
  max_iter: int
    Maximum number of iterations to perform
  """

  def __init__(self, model_builder: Callable[..., Model], max_iter: int):
    super(RandomHyperparamOpt, self).__init__(model_builder=model_builder)
    self.max_iter = max_iter

  def hyperparam_search(
      self,
      params_dict: Dict,
      train_dataset: Dataset,
      valid_dataset: Dataset,
      metric: Metric,
      output_transformers: List[Transformer] = [],
      nb_epoch: int = 10,
      use_max: bool = True,
      logfile: str = 'results.txt',
      logdir: Optional[str] = None,
      **kwargs,
  ) -> Tuple[Model, Dict[str, Any], Dict[str, Any]]:
    """Perform random hyperparams search according to `params_dict`.

    Each key of the `params_dict` is a model_param. The
    values should either be a list of potential values of that hyperparam
    or a callable which can generate random samples.

    Parameters
    ----------
    params_dict: Dict
      Maps each hyperparameter name (string) to either a list of possible
      parameter values or a callable which can generate random samples.
    train_dataset: Dataset
      dataset used for training
    valid_dataset: Dataset
      dataset used for validation (optimization on valid scores)
    metric: Metric
      metric used for evaluation
    output_transformers: list[Transformer]
      Transformers for evaluation. This argument is needed since
      `train_dataset` and `valid_dataset` may have been transformed
      for learning and need the transform to be inverted before
      the metric can be evaluated on a model.
    nb_epoch: int, (default 10)
      Specifies the number of training epochs during each iteration of optimization.
      Not used by all model types.
    use_max: bool, optional
      If True, return the model with the highest score. Else return
      model with the minimum score.
    logdir: str, optional
      The directory in which to store created models. If not set, will
      use a temporary directory.
    logfile: str, optional (default `results.txt`)
      Name of logfile to write results to. If specified, this is must
      be a valid file name. If not specified, results of hyperparameter
      search will be written to `logdir/results.txt`.

    Returns
    -------
    Tuple[`best_model`, `best_hyperparams`, `all_scores`]
      `(best_model, best_hyperparams, all_scores)` where `best_model` is
      an instance of `dc.model.Model`, `best_hyperparams` is a
      dictionary of parameters, and `all_scores` is a dictionary mapping
      string representations of hyperparameter sets to validation
      scores.
    """

    # hyperparam_list should either be an Iterable sequence or a random sampler with rvs method
    for hyperparam in params_dict.values():
      assert isinstance(hyperparam,
                        collections.abc.Iterable) or callable(hyperparam)

    if use_max:
      best_validation_score = -np.inf
    else:
      best_validation_score = np.inf

    best_model = None
    all_scores = {}

    if logdir is not None:
      if not os.path.exists(logdir):
        os.makedirs(logdir, exist_ok=True)
      log_file = os.path.join(logdir, logfile)

    hyperparameter_combs = RandomHyperparamOpt.generate_random_hyperparam_values(
        params_dict, self.max_iter)

    for ind, model_params in enumerate(hyperparameter_combs):
      logger.info("Fitting model %d/%d" % (ind + 1, self.max_iter))
      logger.info("hyperparameters: %s" % str(model_params))

      hp_str = _convert_hyperparam_dict_to_filename(model_params)

      if logdir is not None:
        model_dir = os.path.join(logdir, hp_str)
        logger.info("model_dir is %s" % model_dir)
        try:
          os.makedirs(model_dir)
        except OSError:
          if not os.path.isdir(model_dir):
            logger.info("Error creating model_dir, using tempfile directory")
            model_dir = tempfile.mkdtemp()
      else:
        model_dir = tempfile.mkdtemp()

      model_params['model_dir'] = model_dir
      model = self.model_builder(**model_params)

      # mypy test throws error, so ignoring it in try
      try:
        model.fit(train_dataset, nb_epoch=nb_epoch)  # type: ignore
      # Not all models have nb_epoch
      except TypeError:
        model.fit(train_dataset)
      try:
        model.save()
      # Some models autosave
      except NotImplementedError:
        pass

      multitask_scores = model.evaluate(valid_dataset, [metric],
                                        output_transformers)
      valid_score = multitask_scores[metric.name]

      # Update best validation score so far
      if (use_max and valid_score >= best_validation_score) or (
          not use_max and valid_score <= best_validation_score):
        best_validation_score = valid_score
        best_hyperparams = model_params
        best_model = model
        all_scores[hp_str] = valid_score

      # if `hyp_str` not in `all_scores`, store it in `all_scores`
      if hp_str not in all_scores:
        all_scores[hp_str] = valid_score

      logger.info("Model %d/%d, Metric %s, Validation set %s: %f" %
                  (ind + 1, nb_epoch, metric.name, ind, valid_score))
      logger.info("\tbest_validation_score so far: %f" % best_validation_score)

    if best_model is None:

      logger.info("No models trained correctly.")

      # arbitrarily return last model trained
      if logdir is not None:
        with open(log_file, 'w+') as f:
          f.write("No model trained correctly. Arbitary models returned")

      best_model, best_hyperparams = model, model_params
      return best_model, best_hyperparams, all_scores

    multitask_scores = best_model.evaluate(train_dataset, [metric],
                                           output_transformers)
    train_score = multitask_scores[metric.name]
    logger.info("Best hyperparameters: %s" % str(best_hyperparams))
    logger.info("best train_score: %f" % train_score)
    logger.info("best validation_score: %f" % best_validation_score)

    if logdir is not None:
      with open(log_file, 'w+') as f:
        f.write("Best Hyperparameters dictionary %s\n" % str(best_hyperparams))
        f.write("Best validation score %f\n" % best_validation_score)
        f.write("Best train_score: %f\n" % train_score)
    return best_model, best_hyperparams, all_scores

  @classmethod
  def generate_random_hyperparam_values(cls, params_dict: Dict,
                                        n: int) -> List[Dict[str, Any]]:
    """Generates `n` random hyperparameter combinations of hyperparameter values

    Parameters
    ----------
    params_dict: Dict
      A dictionary of hyperparameters where parameter which takes discrete
      values are specified as iterables and continuous parameters are of
      type callables.
    n: int
      Number of hyperparameter combinations to generate

    Returns
    -------
    A list of generated hyperparameters

    Example
    -------
    >>> from scipy.stats import uniform
    >>> from deepchem.hyper import RandomHyperparamOpt
    >>> n = 1
    >>> params_dict = {'a': [1, 2, 3], 'b': [5, 7, 8], 'c': uniform(10, 5).rvs}
    >>> RandomHyperparamOpt.generate_random_hyperparam_values(params_dict, n)  # doctest: +SKIP
    [{'a': 3, 'b': 7, 'c': 10.619700740985433}]
    """
    hyperparam_keys, hyperparam_values = [], []
    for key, values in params_dict.items():
      if callable(values):
        # If callable, sample it for a maximum n times
        values = [values() for i in range(n)]
      hyperparam_keys.append(key)
      hyperparam_values.append(values)

    hyperparam_combs = []
    for iterable_hyperparam_comb in itertools.product(*hyperparam_values):
      hyperparam_comb = list(iterable_hyperparam_comb)
      hyperparam_combs.append(hyperparam_comb)

    indices = np.random.permutation(len(hyperparam_combs))[:n]
    params_subset = []
    for index in indices:
      param = {}
      for key, hyperparam_value in zip(hyperparam_keys,
                                       hyperparam_combs[index]):
        param[key] = hyperparam_value
      params_subset.append(param)
    return params_subset



"""
Implementation of the Ferminet class in pytorch
"""

from typing import List, Optional, Any, Tuple
# import torch.nn as nn
from rdkit import Chem
import numpy as np
from deepchem.utils.molecule_feature_utils import ALLEN_ELECTRONEGATIVTY
from deepchem.utils.geometry_utils import compute_pairwise_distances

# from deepchem.models.torch_models import TorchModel
# import deepchem.models.optimizers as optim
from deepchem.utils.electron_sampler import ElectronSampler

# TODO look for the loss function(Hamiltonian)


def test_f(x: np.ndarray) -> np.ndarray:
  # dummy function which can be passed as the parameter f. f gives the log probability
  # TODO replace this function with forward pass of the model in future
  return 2 * np.log(np.random.uniform(low=0, high=1.0, size=np.shape(x)[0]))


class Ferminet:
  """A deep-learning based Variational Monte Carlo method for calculating the ab-initio
    solution of a many-electron system.

    This model aims to calculate the ground state energy of a multi-electron system
    using a baseline solution as the Hartree-Fock. An MCMC technique is used to sample
    electrons and DNNs are used to caluclate the square magnitude of the wavefunction,
    in which electron-electron repulsions also are included in the calculation(in the
    form of Jastrow factor envelopes). The model requires only the nucleus' coordinates
    as input.

    This method is based on the following paper:

    Spencer, James S., et al. Better, Faster Fermionic Neural Networks. arXiv:2011.07125,
    arXiv, 13 Nov. 2020. arXiv.org, http://arxiv.org/abs/2011.07125.
    """

  def __init__(
      self,
      nucleon_coordinates: List[List],
      spin: float,
      charge: int,
      seed: Optional[int] = None,
      batch_no: int = 10,
  ):
    """
    Parameters:
    -----------
    nucleon_coordinates:  List[List]
      A list containing nucleon coordinates as the values with the keys as the element's symbol.
    spin: float
      The total spin of the molecule system.
    charge: int
      The total charge of the molecule system.
    seed: int, optional (default None)
      Random seed to use for electron initialization.
    batch_no: int, optional (default 10)
      Number of batches of the electron's positions to be initialized.

    """
    # super(Ferminet, self).__init__()

    self.nucleon_coordinates = nucleon_coordinates
    self.seed = seed
    self.batch_no = batch_no
    self.spin = spin
    self.ion_charge = charge

  def prepare_input_stream(self,) -> Tuple[Any, Any, Any, Any]:
    """Prepares the one-electron and two-electron input stream for the model.

    Returns:
    --------
    one_electron_up: numpy.ndarray
      numpy array containing one-electron coordinates and distances for the up spin electrons.
    one_electron_down: numpy.ndarray
      numpy array containing one-electron coordinates and distances for the down spin electrons
    two_electron_up: numpy.ndarray
      numpy array containing two-electron coordinates and distances for the up spin electrons
    two_electron_down: numpy.ndarray
      numpy array containing two-electron coordinates and distances for the down spin electrons
    """

    no_electrons = []
    nucleons = []
    electronegativity = []

    table = Chem.GetPeriodicTable()
    index = 0
    for i in self.nucleon_coordinates:
      atomic_num = table.GetAtomicNumber(i[0])
      electronegativity.append([index, ALLEN_ELECTRONEGATIVTY[i[0]]])
      no_electrons.append([atomic_num])
      nucleons.append(i[1])
      index += 1

    self.electron_no: np.ndarray = np.array(no_electrons)
    self.charge: np.ndarray = self.electron_no.reshape(
        np.shape(self.electron_no)[0],)
    self.nucleon_pos: np.ndarray = np.array(nucleons)
    electro_neg = np.array(electronegativity)
    self.inter_atom: np.ndarray = compute_pairwise_distances(
        self.nucleon_pos, self.nucleon_pos)

    if np.sum(self.electron_no) < self.ion_charge:
      raise ValueError("Given charge is not initializable")

    # Initialization for ionic molecules
    if self.ion_charge != 0:
      if len(nucleons) == 1:  # for an atom, directly the charge is applied
        self.electron_no[0][0] -= self.ion_charge
      else:  # for a multiatomic molecule, the most electronegative atom gets a charge of -1 and vice versa. The remaining charges are assigned in terms of decreasing(for anionic charge) and increasing(for cationic charge) electronegativity.
        electro_neg = electro_neg[electro_neg[:, 1].argsort()]
        if self.ion_charge > 0:
          for iter in range(self.ion_charge):
            self.electron_no[int(electro_neg[iter][0])][0] -= 1
        else:
          for iter in range(-self.ion_charge):
            self.electron_no[int(electro_neg[-1 - iter][0])][0] += 1

    total_electrons = np.sum(self.electron_no)
    self.up_spin = (total_electrons + 2 * self.spin) // 2
    self.down_spin = (total_electrons - 2 * self.spin) // 2

    self.molecule: ElectronSampler = ElectronSampler(
        batch_no=self.batch_no,
        central_value=self.nucleon_pos,
        seed=self.seed,
        f=test_f,
        steps=1000)  # sample the electrons using the electron sampler
    self.molecule.gauss_initialize_position(
        self.electron_no)  # initialize the position of the electrons

    one_electron_vector = self.molecule.x - self.nucleon_pos

    shape = np.shape(self.molecule.x)
    two_electron_vector = self.molecule.x.reshape([shape[0], 1, shape[1], 3
                                                  ]) - self.molecule.x

    one_electron_vector = one_electron_vector[0, :, :, :]
    two_electron_vector = two_electron_vector[0, :, :, :]

    self.one_electron_distance: np.ndarray = np.linalg.norm(one_electron_vector,
                                                            axis=-1)
    self.two_electron_distance: np.ndarray = np.linalg.norm(two_electron_vector,
                                                            axis=-1)

    # concatenating distance and vectors arrays
    one_shape = np.shape(self.one_electron_distance)
    one_distance = self.one_electron_distance.reshape(1, one_shape[0],
                                                      one_shape[1], 1)
    one_electron = np.block([one_electron_vector, one_distance])
    two_shape = np.shape(self.two_electron_distance)
    two_distance = self.two_electron_distance.reshape(1, two_shape[0],
                                                      two_shape[1], 1)
    two_electron = np.block([two_electron_vector, two_distance])

    one_electron_up = one_electron[:, :self.up_spin, :]
    one_electron_down = one_electron[:, self.up_spin:, :]

    two_electron_up = two_electron[:, :self.up_spin, :]
    two_electron_down = two_electron[:, self.up_spin:, :]

    return one_electron_up, one_electron_down, two_electron_up, two_electron_down



import numpy as np
from flaky import flaky
import pytest
try:
  import tensorflow as tf
  from tensorflow.keras import layers
  has_tensorflow = True
except:
  has_tensorflow = False

import deepchem as dc
from deepchem.data import NumpyDataset


@flaky
@pytest.mark.tensorflow
def test_compute_model_performance_multitask_classifier():
  n_data_points = 20
  n_features = 1
  n_tasks = 2
  n_classes = 2

  X = np.ones(shape=(n_data_points // 2, n_features)) * -1
  X1 = np.ones(shape=(n_data_points // 2, n_features))
  X = np.concatenate((X, X1))
  class_1 = np.array([[0.0, 1.0] for x in range(int(n_data_points / 2))])
  class_0 = np.array([[1.0, 0.0] for x in range(int(n_data_points / 2))])
  y1 = np.concatenate((class_0, class_1))
  y2 = np.concatenate((class_1, class_0))
  y = np.stack([y1, y2], axis=1)
  dataset = NumpyDataset(X, y)

  features = layers.Input(shape=(n_features))
  dense = layers.Dense(n_tasks * n_classes)(features)
  logits = layers.Reshape((n_tasks, n_classes))(dense)
  output = layers.Softmax()(logits)
  keras_model = tf.keras.Model(inputs=features, outputs=[output, logits])
  model = dc.models.KerasModel(
      keras_model,
      dc.models.losses.SoftmaxCrossEntropy(),
      output_types=['prediction', 'loss'],
      learning_rate=0.01,
      batch_size=n_data_points)

  model.fit(dataset, nb_epoch=1000)
  metric = dc.metrics.Metric(
      dc.metrics.roc_auc_score, np.mean, mode="classification")

  scores = model.evaluate_generator(
      model.default_generator(dataset), [metric], per_task_metrics=True)
  scores = list(scores[1].values())
  # Loosening atol to see if tests stop failing sporadically
  assert np.all(np.isclose(scores, [1.0, 1.0], atol=0.50))


@pytest.mark.tensorflow
def test_compute_model_performance_singletask_classifier():
  """Computes model performance on singletask dataset with one-hot label encoding."""
  n_data_points = 20
  n_features = 10

  X = np.ones(shape=(int(n_data_points / 2), n_features)) * -1
  X1 = np.ones(shape=(int(n_data_points / 2), n_features))
  X = np.concatenate((X, X1))
  class_1 = np.array([[0.0, 1.0] for x in range(int(n_data_points / 2))])
  class_0 = np.array([[1.0, 0.0] for x in range(int(n_data_points / 2))])
  y = np.concatenate((class_0, class_1))
  dataset = NumpyDataset(X, y)

  features = layers.Input(shape=(n_features,))
  dense = layers.Dense(2)(features)
  output = layers.Softmax()(dense)
  keras_model = tf.keras.Model(inputs=features, outputs=[output])
  model = dc.models.KerasModel(
      keras_model, dc.models.losses.SoftmaxCrossEntropy(), learning_rate=0.1)

  model.fit(dataset, nb_epoch=1000)
  metric = dc.metrics.Metric(
      dc.metrics.roc_auc_score, np.mean, mode="classification", n_tasks=1)

  scores = model.evaluate_generator(
      model.default_generator(dataset), [metric], per_task_metrics=True)
  scores = list(scores[1].values())
  assert np.isclose(scores, [1.0], atol=0.05)


@pytest.mark.tensorflow
def test_compute_model_performance_multitask_regressor():
  random_seed = 42
  n_data_points = 20
  n_features = 2
  n_tasks = 2
  np.random.seed(seed=random_seed)

  X = np.random.rand(n_data_points, n_features)
  y1 = np.array([0.5 for x in range(n_data_points)])
  y2 = np.array([-0.5 for x in range(n_data_points)])
  y = np.stack([y1, y2], axis=1)
  dataset = NumpyDataset(X, y)

  features = layers.Input(shape=(n_features,))
  dense = layers.Dense(n_tasks)(features)
  keras_model = tf.keras.Model(inputs=features, outputs=[dense])
  model = dc.models.KerasModel(
      keras_model, dc.models.losses.L2Loss(), learning_rate=0.1)

  model.fit(dataset, nb_epoch=1000)
  metric = [
      dc.metrics.Metric(
          dc.metrics.mean_absolute_error, np.mean, mode="regression"),
  ]
  scores = model.evaluate_generator(
      model.default_generator(dataset), metric, per_task_metrics=True)
  scores = list(scores[1].values())
  assert np.all(np.isclose(scores, [0.0, 0.0], atol=1.0))
